from ai_flare_research_assistant.connectors.rpc.flarepy.api.apimeta import api_endpoints

host = "flare-api.flare.network"
# jsonrpc_port = "9650"
protocol = "https"
# jsonrpc_url = "{}://{}:{}".format(jsonrpc_prot, jsonrpc_host, jsonrpc_port)
url = "{}://{}".format(protocol, host)

urls = {k: ("{}" + v).format(url) for k, v in api_endpoints.items()}



import asyncio
import json
import logging
from contextlib import suppress
from typing import Optional, Dict, AsyncIterable, Any

import websockets
from cachetools import TTLCache
from hexbytes import HexBytes
from komoutils.core import PublishQueue, safe_ensure_future
from komoutils.core.time import the_time_now_is
from web3 import Web3
from web3.datastructures import AttributeDict
from web3.exceptions import BlockNotFound
from websockets.exceptions import ConnectionClosed

from aporacle.connectors.watcher.base_watcher import BaseWatcher


class WSNewBlocksWatcher(BaseWatcher):
    MESSAGE_TIMEOUT = 30.0
    PING_TIMEOUT = 10.0

    def __init__(self,
                 w3: Web3,
                 websocket_url: str):
        super().__init__(w3)
        self._network_on = False
        self._nonce: int = 0
        self._current_block_number: int = -1
        self._websocket_url = websocket_url
        self._node_address = None
        self._client: Optional[websockets.WebSocketClientProtocol] = None
        self._fetch_new_blocks_task: Optional[asyncio.Task] = None
        self._block_cache = TTLCache(maxsize=60, ttl=60)
        self._blocks_to_mend: list = []

        self.blocks_publishing_queue: PublishQueue = PublishQueue()

    @property
    def block_number(self) -> int:
        return self._current_block_number

    @property
    def block_cache(self) -> Dict[HexBytes, AttributeDict]:
        cache_dict: Dict[HexBytes, AttributeDict] = dict([(key, self._block_cache[key])
                                                          for key in self._block_cache.keys()])
        return cache_dict

    async def start_network(self):
        if self._fetch_new_blocks_task is not None:
            await self.stop_network()
        else:
            try:
                self._current_block_number = await self.call_async(getattr, self._w3.eth, "block_number")
            except asyncio.CancelledError:
                raise
            except Exception:
                self.log_with_clock(log_level=logging.INFO, msg="Error fetching newest Ethereum block number.",
                                    exc_info=True)
            await self.connect()
            await self.subscribe(["newHeads"])
            self._fetch_new_blocks_task: asyncio.Task = safe_ensure_future(self.fetch_new_blocks_loop())
            self._network_on = True

    async def stop_network(self):
        if self._fetch_new_blocks_task is not None:
            await self.disconnect()
            self._fetch_new_blocks_task.cancel()
            self._fetch_new_blocks_task = None
            self._network_on = False

    async def connect(self):
        try:
            self.log_with_clock(log_level=logging.INFO, msg=f"WS URL - {self._websocket_url}")
            self._client = await websockets.connect(uri=self._websocket_url)
            return self._client
        except Exception as e:
            self.log_with_clock(log_level=logging.ERROR, msg=f"ERROR in connection: {e}")

    async def disconnect(self):
        try:
            await self._client.close()
            self._client = None
        except Exception as e:
            self.log_with_clock(log_level=logging.ERROR, msg=f"ERROR in disconnection: {e}")

    async def _send(self, emit_data) -> int:
        if self._client is None:
            while True:
                await self.connect()
                if self._client is not None:
                    break
                else:
                    self.log_with_clock(log_level=logging.ERROR, msg=f"Websocket client has a value of NONE. "
                                                                     f"Will attempt to reconnect in 10 seconds. ")
                    await asyncio.sleep(10)

        self._nonce += 1
        emit_data["id"] = self._nonce
        await self._client.send(json.dumps(emit_data))
        return self._nonce

    async def subscribe(self, params) -> bool:
        emit_data = {
            "method": "eth_subscribe",
            "params": params
        }
        nonce = await self._send(emit_data)
        raw_message = await self._client.recv()
        if raw_message is not None:
            resp = json.loads(raw_message)
            if resp.get("id", None) == nonce:
                self._node_address = resp.get("result")
                return True
        return False

    async def _messages(self) -> AsyncIterable[Any]:
        try:
            while True:
                try:
                    raw_msg_str: str = await asyncio.wait_for(self._client.recv(), self.MESSAGE_TIMEOUT)
                    yield raw_msg_str
                except asyncio.TimeoutError:
                    pong_waiter = await self._client.ping()
                    await asyncio.wait_for(pong_waiter, timeout=self.PING_TIMEOUT)
        except asyncio.TimeoutError:
            self.log_with_clock(log_level=logging.WARNING, msg="WebSocket ping timed out. Going to reconnect...")
            return
        except ConnectionClosed:
            return
        finally:
            await self.disconnect()
            # Reconnect and subscribe in case a disconnect happens
            await self.connect()
            await self.subscribe(["newHeads"])

    async def fetch_new_blocks_loop(self):
        while True:
            try:
                async for raw_message in self._messages():
                    message_json = json.loads(raw_message) if raw_message is not None else None
                    if message_json.get("method", None) == "eth_subscription":
                        subscription_result_params = message_json.get("params", None)
                        incoming_block = subscription_result_params.get("result", None) \
                            if subscription_result_params is not None else None
                        if incoming_block is not None:
                            with suppress(BlockNotFound):
                                try:
                                    new_block: AttributeDict = await self.call_async(self._w3.eth.get_block,
                                                                                     incoming_block.get("hash"), True)
                                    self._current_block_number = new_block.get("number")
                                    # print(new_block.get("hash"))
                                    self.log_with_clock(log_level=logging.DEBUG,
                                                        msg=f"Block number - {self._current_block_number}. ")
                                    self._block_cache[new_block.get("hash")] = new_block
                                    # print(new_block)
                                    self.blocks_publishing_queue.publish([new_block])
                                    # self.trigger_event(NewBlocksWatcherEvent.NewBlocks, [new_block])
                                except Exception as e:
                                    self.log_with_clock(log_level=logging.ERROR, msg=f"ERROR - {e}")

            except asyncio.TimeoutError:
                self.log_with_clock(log_level=logging.ERROR, msg="Timed out fetching new block.", exc_info=True)
                await asyncio.sleep(30.0)
                await self.start_network()
            except asyncio.CancelledError as e:
                self.log_with_clock(log_level=logging.ERROR, msg=f"A Cancel Error fetching new block: {e}",
                                    exc_info=True)
                await asyncio.sleep(30.0)
                await self.start_network()
            except Exception as e:
                self.log_with_clock(log_level=logging.ERROR, msg=f"Error fetching new block: {e}", exc_info=True)
                await asyncio.sleep(5.0)
                await self.start_network()

    async def get_timestamp_for_block(self, block_hash: HexBytes, max_tries: Optional[int] = 10) -> int:
        counter = 0
        block: AttributeDict = None
        if block_hash in self._block_cache.keys():
            block = self._block_cache.get(block_hash)
        else:
            return int(the_time_now_is())
            # while block is None:
            #     if counter == max_tries:
            #         raise ValueError(f"Block hash {block_hash.hex()} does not exist.")
            #     counter += 1
            #     block = self._block_cache.get(block_hash)
            #     await asyncio.sleep(0.5)
        return block.get("timestamp")



from typing import Any, Dict, Optional

from eth_account import Account
from eth_account.messages import defunct_hash_message
from eth_account.signers.local import LocalAccount
from komoutils.core import KomoBase
from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.exceptions import BlockNotFound
from web3.middleware import geth_poa_middleware

from chain_data_collector.chain.evm_chain import FlareChain
from chain_data_collector.chain.watcher.websocket_watcher import WSNewBlocksWatcher
from chain_data_collector.core.utils.async_call_scheduler import AsyncCallScheduler


class OracleWallet(KomoBase):
    def __init__(self,
                 private_key: Any,
                 evm_rpc_url: str,
                 evm_chain_ws: str,
                 chain: FlareChain = FlareChain.SONGBIRD):
        super().__init__()

        # Initialize Web3, accounts and contracts.
        self._evm_rpc_url: str = evm_rpc_url
        self._evm_ws_url: str = evm_chain_ws

        self._w3: Web3 = Web3(Web3.HTTPProvider(evm_rpc_url))
        self._w3.middleware_onion.inject(geth_poa_middleware, layer=0)

        self._chain: FlareChain = chain
        self._account: LocalAccount = Account.from_key(private_key)

        self._new_blocks_watcher: Optional[WSNewBlocksWatcher] = None


    @property
    def address(self) -> str:
        return self._account.address

    @property
    def get_block(self):
        block = self._w3.eth.get_block('latest')
        return block

    @property
    def block_number(self) -> int:
        return self._new_blocks_watcher.block_number if self._new_blocks_watcher is not None else -1

    @property
    def gas_price(self) -> int:
        """
        Warning: This property causes network access, even though it's synchronous.

        :return: Gas price in wei
        """
        # TODO: The gas price from Parity is not reliable. Convert to use internal gas price calculator
        return self._gas_price

    @property
    def nonce(self) -> int:
        """
        Warning: This property causes network access, even though it's synchronous.

        :return: Gas price in wei
        """
        remote_nonce: int = self.get_remote_nonce()
        retval: int = max(remote_nonce, self._local_nonce)
        self._local_nonce = retval
        return retval

    @property
    def chain(self) -> FlareChain:
        return self._chain

    def sign_hash(self, text: str = None, hexstr: str = None) -> str:
        msg_hash: str = defunct_hash_message(hexstr=hexstr, text=text)
        signature_dict: AttributeDict = self._account.signHash(msg_hash)
        signature: str = signature_dict["signature"].hex()
        return signature

    def execute_transaction(self, contract_function: ContractFunction, **kwargs) -> str:
        """
        This function WILL result in immediate network calls (e.g. to get the gas price, nonce and gas cost), even
        though it is written in sync manner.

        :param contract_function:
        :param kwargs:
        :return:
        """
        # if self._network_status is not NetworkStatus.CONNECTED:
        #     raise EnvironmentError("Cannot send transactions when network status is not connected.")

        # print(f"BACKEND ADDRESS {self.address}")
        gas_price: int = self.gas_price
        print(f"Submitted Gas price {gas_price}")
        transaction_args: Dict[str, Any] = {
            "from": self.address,
            "nonce": self.nonce,
            # "chainId": self.chain.value,
            "gasPrice": gas_price
        }
        transaction_args.update(kwargs)
        transaction: Dict[str, Any] = contract_function.buildTransaction(transaction_args)
        if "gas" not in transaction:
            estimate_gas: int = 1000000
            try:
                estimate_gas = self._w3.eth.estimateGas(transaction)
            except ValueError:
                self.logger().error("Failed to estimate gas. Using default of 1000000.")
            transaction["gas"] = estimate_gas
        signed_transaction: AttributeDict = self._account.signTransaction(transaction)
        tx_hash: str = signed_transaction.hash.hex()
        self.schedule_eth_transaction(signed_transaction, gas_price)
        return tx_hash

    async def _update_gas_price(self):
        async_scheduler: AsyncCallScheduler = AsyncCallScheduler.shared_instance()
        new_gas_price: int = await async_scheduler.call_async(getattr, self._w3.eth, "gasPrice")
        self._gas_price = new_gas_price

    def get_remote_nonce(self):
        try:
            remote_nonce = self._w3.eth.getTransactionCount(self.address, block_identifier="pending")
            return remote_nonce
        except BlockNotFound:
            return None



import json
import os

from web3 import Web3

from aporacle.connectors.oracle.contracts.tso_contract_base import TsoContractBase


class FtsoFeedPublisherContract(TsoContractBase):

    def __init__(self,
                 w3: Web3,
                 address: str):
        super().__init__(w3, address)

    @property
    def name(self):
        return 'ftso_feed_publisher_contract'

    @property
    def abi(self):
        chain = os.getenv("CHAIN")
        with open(
                os.path.join(os.path.dirname(__file__), f'{self.name}_abi.{chain}.json')) as contract_abi:
            data: dict = json.load(contract_abi)
        return data["abi"]

    def get_feed(self, name: str):
        return self.contract.functions.getFeed(name).call()



import asyncio
import logging
from typing import List, Coroutine, Union, Optional

from komoutils.core import safe_gather
from komoutils.logger import KomoLogger
from web3 import Web3
from web3.contract import Contract

from aporacle.core.utils.async_call_scheduler import AsyncCallScheduler


class TsoContractBase:
    _tso_contract_logger: Optional[KomoLogger] = None

    @classmethod
    def logger(cls) -> KomoLogger:
        if cls._tso_contract_logger is None:
            cls._tso_contract_logger = logging.getLogger(__name__)
        return cls._tso_contract_logger

    def __init__(self,
                 w3: Web3,
                 address: str):
        self._w3: Web3 = w3
        self._address: str = address
        self._name: str = ""
        self._contract: Contract = self._w3.eth.contract(address=self._address, abi=self.abi)

    @property
    def name(self):
        return NotImplementedError

    @property
    def abi(self):
        return NotImplementedError

    @classmethod
    def get_name_from_contract(cls, contract: Contract) -> str:
        raw_name: Union[str, bytes] = contract.functions.name().call()
        if isinstance(raw_name, bytes):
            retval: str = raw_name.split(b"\x00")[0].decode("utf8")
        else:
            retval: str = raw_name
        return retval

    @property
    def address(self) -> str:
        return self._address

    @property
    def contract(self) -> Contract:
        return self._contract

    async def _get_contract_info(self):
        if self._name is not None:
            return

        tasks: List[Coroutine] = [
            AsyncCallScheduler.shared_instance().call_async(func, *args)
            for func, args in [
                (self.get_name_from_contract, [self._contract]),
            ]
        ]

        try:
            name = await safe_gather(*tasks)
            self._name = name
            # self._decimals = decimals
        except asyncio.CancelledError as e:
            raise e
        except Exception:
            self.logger().network(f"Error fetching price contract info for {self._contract.address}.",
                                  exc_info=True,
                                  app_warning_msg=f"Error fetching price contract info for {self._contract.address}. "
                                                  f"Check wallet network connection")

    async def get_name(self) -> str:
        if self._name is None:
            await self._get_contract_info()
        return self._name



import logging
from typing import List, Optional

import pandas as pd
from komoutils.core import KomoBase

from aporacle.data.asset import AssetData
from aporacle.data.gcp import download_csv_from_gcp_return_df
from aporacle.data.symbols import SymbolData


class Combination(KomoBase):
    def __init__(self, feed: str, symbols: List[str] = None):
        self.feed: str = feed
        self.symbols: Optional[list] = symbols
        self.combined: Optional[pd.DataFrame] = None

    def combine(self, data: List[pd.DataFrame]):
        self.combined = pd.concat(data, axis=1)
        return self.combined

    def get_from_gcp(self):
        try:
            assert len(self.symbols) > 0, f"Please provide valid symbols. "
            data: List[pd.DataFrame] = [download_csv_from_gcp_return_df(bucket_name=self.feed, symbol=symbol)
                                        for symbol in self.symbols]
            return self.combine(data)
        except AssertionError as ae:
            self.log_with_clock(log_level=logging.ERROR, msg=f"{ae}")
        except Exception as e:
            raise

    def get_from_provided(self, dfs: List[pd.DataFrame]):
        try:
            data: List[pd.DataFrame] = [df for df in dfs]
            return self.combine(data)
        except Exception as e:
            raise


#!/usr/bin/env python

import asyncio
from typing import (
    Coroutine,
    Callable
)

from komoutils.core import KomoBase
from web3 import Web3

from aporacle.core.utils.async_call_scheduler import AsyncCallScheduler


class BaseWatcher(KomoBase):
    def __init__(self, w3: Web3):
        super().__init__()
        self._w3: Web3 = w3
        self._ev_loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()

    @staticmethod
    async def schedule_async_call(coro: Coroutine, timeout_seconds: float, **kwargs) -> any:
        return await AsyncCallScheduler.shared_instance().schedule_async_call(coro, timeout_seconds, **kwargs)

    @staticmethod
    async def call_async(func: Callable, *args, **kwargs):
        return await AsyncCallScheduler.shared_instance().call_async(func, *args, **kwargs)

    async def start_network(self):
        raise NotImplementedError

    async def stop_network(self):
        raise NotImplementedError



api_endpoints = {'admin': '/ext/admin',
                 'avm': '/ext/bc/X',
                 'health': '/ext/health',
                 'info': '/ext/info',
                 'ipcs': '/ext/ipcs',
                 'keystore': '/ext/keystore',
                 'platform': '/ext/bc/P',
                 'timestamp': '/ext/timestamp'}

api_meta = {'/ext/admin': {'admin': {'alias': {'parameters': {'alias': {'optional': False,
                                                                        'type': 'string'},
                                                              'endpoint': {'optional': False,
                                                                           'type': 'string'}},
                                               'returns': {'success': {'optional': False,
                                                                       'type': 'bool'}}},
                                     'aliasChain': {'parameters': {'alias': {'optional': False,
                                                                             'type': 'string'},
                                                                   'chain': {'optional': False,
                                                                             'type': 'string'}},
                                                    'returns': {'success': {'optional': False,
                                                                            'type': 'bool'}}},
                                     'getBlockchainID': {'parameters': {'alias': {'optional': False,
                                                                                  'type': 'string'}},
                                                         'returns': {'blockchainID': {'optional': False,
                                                                                      'type': 'string'}}},
                                     'getNetworkID': {'parameters': {},
                                                      'returns': {'networkID': {'optional': False,
                                                                                'type': 'int'}}},
                                     'getNetworkName': {'parameters': {},
                                                        'returns': {'networkName': {'optional': False,
                                                                                    'type': 'string'}}},
                                     'getNodeID': {'parameters': {},
                                                   'returns': {'nodeID': {'optional': False,
                                                                          'type': 'string'}}},
                                     'getNodeVersion': {'parameters': {},
                                                        'returns': {'version': {'optional': False,
                                                                                'type': 'string'}}},
                                     'lockProfile': {'parameters': {'fileName': {'optional': False,
                                                                                 'type': 'string'}},
                                                     'returns': {'success': {'optional': False,
                                                                             'type': 'bool'}}},
                                     'memoryProfile': {'parameters': {'fileName': {'optional': False,
                                                                                   'type': 'string'}},
                                                       'returns': {'success': {'optional': False,
                                                                               'type': 'bool'}}},
                                     'peers': {'parameters': {},
                                               'returns': {'peers': {'optional': False,
                                                                     'type': {'[]': {}}}}},
                                     'startCPUProfiler': {'parameters': {'fileName': {'optional': False,
                                                                                      'type': 'string'}},
                                                          'returns': {'success': {'optional': False,
                                                                                  'type': 'bool'}}},
                                     'stopCPUProfiler': {'parameters': {},
                                                         'returns': {'success': {'optional': False,
                                                                                 'type': 'bool'}}}}},
            '/ext/bc/P': {'platform': {'addDefaultSubnetDelegator': {'parameters': {'destination': {'optional': False,
                                                                                                    'type': 'string'},
                                                                                    'endTime': {'optional': False,
                                                                                                'type': 'int'},
                                                                                    'id': {'optional': False,
                                                                                           'type': 'string'},
                                                                                    'payerNonce': {'optional': False,
                                                                                                   'type': 'int'},
                                                                                    'stakeAmount': {'optional': False,
                                                                                                    'type': 'int'},
                                                                                    'startTime': {'optional': False,
                                                                                                  'type': 'int'}},
                                                                     'returns': {'unsignedTx': {'optional': False,
                                                                                                'type': 'string'}}},
                                       'addDefaultSubnetValidator': {
                                           'parameters': {'delegationFeeRate': {'optional': False,
                                                                                'type': 'int'},
                                                          'destination': {'optional': False,
                                                                          'type': 'string'},
                                                          'endTime': {'optional': False,
                                                                      'type': 'int'},
                                                          'id': {'optional': False,
                                                                 'type': 'string'},
                                                          'payerNonce': {'optional': False,
                                                                         'type': 'int'},
                                                          'stakeAmount': {'optional': False,
                                                                          'type': 'int'},
                                                          'startTime': {'optional': False,
                                                                        'type': 'int'}},
                                           'returns': {'unsignedTx': {'optional': False,
                                                                      'type': 'string'}}},
                                       'addNonDefaultSubnetValidator': {'parameters': {'endTime': {'optional': False,
                                                                                                   'type': 'int'},
                                                                                       'id': {'optional': False,
                                                                                              'type': 'string'},
                                                                                       'payerNonce': {'optional': False,
                                                                                                      'type': 'int'},
                                                                                       'startTime': {'optional': False,
                                                                                                     'type': 'int'},
                                                                                       'subnetID': {'optional': False,
                                                                                                    'type': 'string'},
                                                                                       'weight': {'optional': False,
                                                                                                  'type': 'int'}},
                                                                        'returns': {'unsignedTx': {'optional': False,
                                                                                                   'type': 'string'}}},
                                       'createAccount': {'parameters': {'password': {'optional': False,
                                                                                     'type': 'string'},
                                                                        'username': {'optional': False,
                                                                                     'type': 'string'}},
                                                         'returns': {'address': {'optional': False,
                                                                                 'type': 'string'}}},
                                       'createBlockchain': {'parameters': {'genesisData': {'optional': False,
                                                                                           'type': 'string'},
                                                                           'name': {'optional': False,
                                                                                    'type': 'string'},
                                                                           'payerNonce': {'optional': False,
                                                                                          'type': 'int'},
                                                                           'subnetID': {'optional': False,
                                                                                        'type': 'string'},
                                                                           'vmID': {'optional': False,
                                                                                    'type': 'string'}},
                                                            'returns': {'unsignedTx': {'optional': False,
                                                                                       'type': 'string'}}},
                                       'createSubnet': {'parameters': {'controlKeys': {'optional': False,
                                                                                       'type': {'[]': {}}},
                                                                       'payerNonce': {'optional': False,
                                                                                      'type': 'int'},
                                                                       'threshold': {'optional': False,
                                                                                     'type': 'int'}},
                                                        'returns': {'unsignedTx': {'optional': False,
                                                                                   'type': 'string'}}},
                                       'exportAVA': {'parameters': {'amount': {'optional': False,
                                                                               'type': 'int'},
                                                                    'payerNonce': {'optional': False,
                                                                                   'type': 'int'},
                                                                    'to': {'optional': False,
                                                                           'type': 'string'}},
                                                     'returns': {'unsigndTx': {'optional': False,
                                                                               'type': 'string'}}},
                                       'exportKey': {'parameters': {'address': {'optional': False,
                                                                                'type': 'string'},
                                                                    'password': {'optional': False,
                                                                                 'type': 'string'},
                                                                    'username': {'optional': False,
                                                                                 'type': 'string'}},
                                                     'returns': {'privateKey': {'optional': False,
                                                                                'type': 'string'}}},
                                       'getAccount': {'parameters': {'address': {'optional': False,
                                                                                 'type': 'string'}},
                                                      'returns': {'address': {'optional': False,
                                                                              'type': 'string'},
                                                                  'balance': {'optional': False,
                                                                              'type': 'int'},
                                                                  'nonce': {'optional': False,
                                                                            'type': 'int'}}},
                                       'getBlockchainStatus': {'parameters': {'blockchainID': {'optional': False,
                                                                                               'type': 'string'}},
                                                               'returns': {'status': {'optional': False,
                                                                                      'type': 'string'}}},
                                       'getBlockchains': {'parameters': {},
                                                          'returns': {'blockchains': {'optional': False,
                                                                                      'type': {'[]': {}}}}},
                                       'getCurrentValidators': {'parameters': {'subnetID': {'optional': False,
                                                                                            'type': 'string'}},
                                                                'returns': {'validators': {'optional': False,
                                                                                           'type': {'[]': {}}}}},
                                       'getPendingValidators': {'parameters': {'subnetID': {'optional': False,
                                                                                            'type': 'string'}},
                                                                'returns': {'validators': {'optional': False,
                                                                                           'type': {'[]': {}}}}},
                                       'getSubnets': {'parameters': {},
                                                      'returns': {'subnets': {'optional': False,
                                                                              'type': {'[]': {}}}}},
                                       'importAVA': {'parameters': {'password': {'optional': False,
                                                                                 'type': 'string'},
                                                                    'payerNonce': {'optional': False,
                                                                                   'type': 'int'},
                                                                    'to': {'optional': False,
                                                                           'type': 'string'},
                                                                    'username': {'optional': False,
                                                                                 'type': 'string'}},
                                                     'returns': {'tx': {'optional': False,
                                                                        'type': 'string'}}},
                                       'importKey': {'parameters': {'password': {'optional': False,
                                                                                 'type': 'string'},
                                                                    'privateKey': {'optional': False,
                                                                                   'type': 'string'},
                                                                    'username': {'optional': False,
                                                                                 'type': 'string'}},
                                                     'returns': {'address': {'optional': False,
                                                                             'type': 'string'}}},
                                       'issueTx': {'parameters': {'tx': {'optional': False,
                                                                         'type': 'string'}},
                                                   'returns': {'txID': {'optional': False,
                                                                        'type': 'string'}}},
                                       'listAccounts': {'parameters': {'password': {'optional': False,
                                                                                    'type': 'string'},
                                                                       'username': {'optional': False,
                                                                                    'type': 'string'}},
                                                        'returns': {'accounts': {'optional': False,
                                                                                 'type': {'[]': {}}}}},
                                       'sampleValidators': {'parameters': {'size': {'optional': False,
                                                                                    'type': 'int'},
                                                                           'subnetID': {'optional': False,
                                                                                        'type': 'string'}},
                                                            'returns': {'validators': {'optional': False,
                                                                                       'type': '[size]string'}}},
                                       'sign': {'parameters': {'password': {'optional': False,
                                                                            'type': 'string'},
                                                               'signer': {'optional': False,
                                                                          'type': 'string'},
                                                               'tx': {'optional': False,
                                                                      'type': 'string'},
                                                               'username': {'optional': False,
                                                                            'type': 'string'}},
                                                'returns': {'tx': {'optional': False,
                                                                   'type': 'string'}}},
                                       'validatedBy': {'parameters': {'blockchainID': {'optional': False,
                                                                                       'type': 'string'}},
                                                       'returns': {'subnetID': {'optional': False,
                                                                                'type': 'string'}}},
                                       'validates': {'parameters': {'subnetID': {'optional': False,
                                                                                 'type': 'string'}},
                                                     'returns': {'blockchainIDs': {'optional': False,
                                                                                   'type': {'[]': {'optional': False,
                                                                                                   'type': 'string'}}}}}}},
            '/ext/bc/X': {'avm': {'Send': {'parameters': {'amount': {'optional': False,
                                                                     'type': 'int'},
                                                          'assetID': {'optional': False,
                                                                      'type': 'string'},
                                                          'password': {'optional': False,
                                                                       'type': 'string'},
                                                          'to': {'optional': False,
                                                                 'type': 'string'},
                                                          'username': {'optional': False,
                                                                       'type': 'string'}},
                                           'returns': {'txID': {'optional': False,
                                                                'type': 'string'}}},
                                  'buildGenesis': {'parameters': {'genesisData': {'optional': False,
                                                                                  'type': 'JSON'}},
                                                   'returns': {'bytes': {'optional': False,
                                                                         'type': 'string'}}},
                                  'createAddress': {'parameters': {'password': {'optional': False,
                                                                                'type': 'string'},
                                                                   'username': {'optional': False,
                                                                                'type': 'string'}},
                                                    'returns': {'address': {'optional': False,
                                                                            'type': 'string'}}},
                                  'createFixedCapAsset': {'parameters': {'denomination': {'optional': False,
                                                                                          'type': 'int'},
                                                                         'initialHolders': {'optional': False,
                                                                                            'type': {'[]': {}}},
                                                                         'name': {'optional': False,
                                                                                  'type': 'string'},
                                                                         'password': {'optional': False,
                                                                                      'type': 'string'},
                                                                         'symbol': {'optional': False,
                                                                                    'type': 'string'},
                                                                         'username': {'optional': False,
                                                                                      'type': 'string'}},
                                                          'returns': {'assetID': {'optional': False,
                                                                                  'type': 'string'}}},
                                  'createMintTx': {'parameters': {'amount': {'optional': False,
                                                                             'type': 'int'},
                                                                  'assetID': {'optional': False,
                                                                              'type': 'string'},
                                                                  'minters': {'optional': False,
                                                                              'type': {'[]': {'optional': False,
                                                                                              'type': 'string'}}},
                                                                  'to': {'optional': False,
                                                                         'type': 'string'}},
                                                   'returns': {'tx': {'optional': False,
                                                                      'type': 'string'}}},
                                  'createVariableCapAsset': {'parameters': {'denomination': {'optional': False,
                                                                                             'type': 'int'},
                                                                            'minterSets': {'optional': False,
                                                                                           'type': {'[]': {}}},
                                                                            'name': {'optional': False,
                                                                                     'type': 'string'},
                                                                            'password': {'optional': False,
                                                                                         'type': 'string'},
                                                                            'symbol': {'optional': False,
                                                                                       'type': 'string'},
                                                                            'username': {'optional': False,
                                                                                         'type': 'string'}},
                                                             'returns': {'assetID': {'optional': False,
                                                                                     'type': 'string'}}},
                                  'exportAVA': {'parameters': {'amount': {'optional': False,
                                                                          'type': 'int'},
                                                               'password': {'optional': False,
                                                                            'type': 'string'},
                                                               'to': {'optional': False,
                                                                      'type': 'string'},
                                                               'username': {'optional': False,
                                                                            'type': 'string'}},
                                                'returns': {'txID': {'optional': False,
                                                                     'type': 'string'}}},
                                  'exportKey': {'parameters': {'address': {'optional': False,
                                                                           'type': 'string'},
                                                               'password': {'optional': False,
                                                                            'type': 'string'},
                                                               'username': {'optional': False,
                                                                            'type': 'string'}},
                                                'returns': {'privateKey': {'optional': False,
                                                                           'type': 'string'}}},
                                  'getAllBalances': {'parameters': {'address': {'optional': False,
                                                                                'type': 'string'}},
                                                     'returns': {'balances': {'optional': False,
                                                                              'type': {'[]': {}}}}},
                                  'getAssetDescription': {'parameters': {'assetID': {'optional': False,
                                                                                     'type': 'string'}},
                                                          'returns': {'denomination': {'optional': False,
                                                                                       'type': 'int'},
                                                                      'name': {'optional': False,
                                                                               'type': 'string'},
                                                                      'symbol': {'optional': False,
                                                                                 'type': 'string'}}},
                                  'getBalance': {'parameters': {'address': {'optional': False,
                                                                            'type': 'string'},
                                                                'assetID': {'optional': False,
                                                                            'type': 'string'}},
                                                 'returns': {'balance': {'optional': False,
                                                                         'type': 'int'}}},
                                  'getTx': {'parameters': {'txID': {'optional': False,
                                                                    'type': 'string'}},
                                            'returns': {'tx': {'optional': False,
                                                               'type': 'string'}}},
                                  'getTxStatus': {'parameters': {'txID': {'optional': False,
                                                                          'type': 'string'}},
                                                  'returns': {'status': {'optional': False,
                                                                         'type': 'string'}}},
                                  'getUTXOs': {'parameters': {'addresses': {'optional': False,
                                                                            'type': 'string'}},
                                               'returns': {'utxos': {'optional': False,
                                                                     'type': {'[]': {'optional': False,
                                                                                     'type': 'string'}}}}},
                                  'importAVA': {'parameters': {'password': {'optional': False,
                                                                            'type': 'string'},
                                                               'to': {'optional': False,
                                                                      'type': 'string'},
                                                               'username': {'optional': False,
                                                                            'type': 'string'}},
                                                'returns': {'txID': {'optional': False,
                                                                     'type': 'string'}}},
                                  'importKey': {'parameters': {'password': {'optional': False,
                                                                            'type': 'string'},
                                                               'privateKey': {'optional': False,
                                                                              'type': 'string'},
                                                               'username': {'optional': False,
                                                                            'type': 'string'}},
                                                'returns': {'address': {'optional': False,
                                                                        'type': 'string'}}},
                                  'issueTx': {'parameters': {'tx': {'optional': False,
                                                                    'type': 'string'}},
                                              'returns': {'txID': {'optional': False,
                                                                   'type': 'string'}}},
                                  'listAddresses': {'parameters': {'password': {'optional': False,
                                                                                'type': 'string'},
                                                                   'username': {'optional': False,
                                                                                'type': 'string'}},
                                                    'returns': {'addresses': {'optional': False,
                                                                              'type': {'[]': {'optional': False,
                                                                                              'type': 'string'}}}}},
                                  'signMintTx': {'parameters': {'minter': {'optional': False,
                                                                           'type': 'string'},
                                                                'password': {'optional': False,
                                                                             'type': 'string'},
                                                                'tx': {'optional': False,
                                                                       'type': 'string'},
                                                                'username': {'optional': False,
                                                                             'type': 'string'}},
                                                 'returns': {'tx': {'optional': False,
                                                                    'type': 'string'}}}}},
            '/ext/health': {'health': {'getLiveness': {'parameters': {},
                                                       'returns': {'checks': {'optional': False,
                                                                              'type': {'{}': {}}},
                                                                   'healthy': {'optional': False,
                                                                               'type': 'bool'}}}}},
            '/ext/info': {'info': {'getBlockchainID': {'parameters': {'alias': {'optional': False,
                                                                                'type': 'string'}},
                                                       'returns': {'blockchainID': {'optional': False,
                                                                                    'type': 'string'}}},
                                   'getNetworkID': {'parameters': {},
                                                    'returns': {'networkID': {'optional': False,
                                                                              'type': 'int'}}},
                                   'getNetworkName': {'parameters': {},
                                                      'returns': {'networkName': {'optional': False,
                                                                                  'type': 'string'}}},
                                   'getNodeID': {'parameters': {},
                                                 'returns': {'nodeID': {'optional': False,
                                                                        'type': 'string'}}},
                                   'getNodeVersion': {'parameters': {},
                                                      'returns': {'version': {'optional': False,
                                                                              'type': 'string'}}},
                                   'peers': {'parameters': {},
                                             'returns': {'peers': {'optional': False,
                                                                   'type': {'[]': {}}}}}}},
            '/ext/ipcs': {'ipcs': {'publishBlockchain': {'parameters': {'blockchainID': {'optional': False,
                                                                                         'type': 'string'}},
                                                         'returns': {'url': {'optional': False,
                                                                             'type': 'string'}}},
                                   'unpublishBlockchain': {'parameters': {'blockchainID': {'optional': False,
                                                                                           'type': 'string'}},
                                                           'returns': {'success': {'optional': False,
                                                                                   'type': 'bool'}}}}},
            '/ext/keystore': {'keystore': {'ListUsers': {'parameters': {},
                                                         'returns': {'users': {'optional': False,
                                                                               'type': {'[]': {'optional': False,
                                                                                               'type': 'string'}}}}},
                                           'createUser': {'parameters': {'password': {'optional': False,
                                                                                      'type': 'string'},
                                                                         'username': {'optional': False,
                                                                                      'type': 'string'}},
                                                          'returns': {'success': {'optional': False,
                                                                                  'type': 'bool'}}},
                                           'deleteUser': {'parameters': {'password': {'optional': False,
                                                                                      'type': 'string'},
                                                                         'username': {'optional': False,
                                                                                      'type': 'string'}},
                                                          'returns': {'success': {'optional': False,
                                                                                  'type': 'bool'}}},
                                           'exportUser': {'parameters': {'password': {'optional': False,
                                                                                      'type': 'string'},
                                                                         'username': {'optional': False,
                                                                                      'type': 'string'}},
                                                          'returns': {'user': {'optional': False,
                                                                               'type': 'string'}}},
                                           'importUser': {'parameters': {'password': {'optional': False,
                                                                                      'type': 'string'},
                                                                         'user': {'optional': False,
                                                                                  'type': 'string'},
                                                                         'username': {'optional': False,
                                                                                      'type': 'string'}},
                                                          'returns': {'success': {'optional': False,
                                                                                  'type': 'bool'}}}}},
            '/ext/timestamp': {'timestamp': {'getBlock': {'parameters': {'id': {'optional': False,
                                                                                'type': 'string'}},
                                                          'returns': {'data': {'optional': False,
                                                                               'type': 'string'},
                                                                      'id': {'optional': False,
                                                                             'type': 'string'},
                                                                      'parentID': {'optional': False,
                                                                                   'type': 'string'},
                                                                      'timestamp': {'optional': False,
                                                                                    'type': 'int'}}},
                                             'proposeBlock': {'parameters': {'data': {'optional': False,
                                                                                      'type': 'string'}},
                                                              'returns': {'success': {'optional': False,
                                                                                      'type': 'bool'}}}}}}
